@@ -17,20 +17,22 @@ import (
 	"time"
 )
 
-var (
-	killTime    = 500 * time.Millisecond
-	timeoutTime = 1000 * time.Millisecond
-	waitTime    = 100 * time.Millisecond
+const (
+	// The tests will run a test server on this port.
+	port               = 9654
+	concurrentRequestN = 8
+	killTime           = 500 * time.Millisecond
+	timeoutTime        = 1000 * time.Millisecond
+	waitTime           = 100 * time.Millisecond
 )
 
 func runQuery(t *testing.T, expected int, shouldErr bool, wg *sync.WaitGroup, once *sync.Once) {
-	wg.Add(1)
 	defer wg.Done()
 	client := http.Client{}
-	r, err := client.Get("http://localhost:9654")
+	r, err := client.Get(fmt.Sprintf("http://localhost:%d", port))
 	if shouldErr && err == nil {
 		once.Do(func() {
-			t.Fatal("Expected an error but none was encountered.")
+			t.Error("Expected an error but none was encountered.")
 		})
 	} else if shouldErr && err != nil {
 		if checkErr(t, err, once) {
@@ -39,11 +41,11 @@ func runQuery(t *testing.T, expected int, shouldErr bool, wg *sync.WaitGroup, on
 	}
 	if r != nil && r.StatusCode != expected {
 		once.Do(func() {
-			t.Fatalf("Incorrect status code on response. Expected %d. Got %d", expected, r.StatusCode)
+			t.Errorf("Incorrect status code on response. Expected %d. Got %d", expected, r.StatusCode)
 		})
 	} else if r == nil {
 		once.Do(func() {
-			t.Fatal("No response when a response was expected.")
+			t.Error("No response when a response was expected.")
 		})
 	}
 }
@@ -65,13 +67,16 @@ func checkErr(t *testing.T, err error, once *sync.Once) bool {
 			return true
 		} else if err != nil {
 			once.Do(func() {
-				t.Fatal("Error on Get:", err)
+				t.Error("Error on Get:", err)
 			})
 		}
 	default:
 		if strings.Contains(err.Error(), "transport closed before response was received") {
 			return true
 		}
+		if strings.Contains(err.Error(), "server closed connection") {
+			return true
+		}
 		fmt.Printf("unknown err: %s, %#v\n", err, err)
 	}
 	return false
@@ -84,16 +89,17 @@ func createListener(sleep time.Duration) (*http.Server, net.Listener, error) {
 		rw.WriteHeader(http.StatusOK)
 	})
 
-	server := &http.Server{Addr: ":9654", Handler: mux}
-	l, err := net.Listen("tcp", ":9654")
-	if err != nil {
-	}
+	server := &http.Server{Addr: fmt.Sprintf(":%d", port), Handler: mux}
+	l, err := net.Listen("tcp", fmt.Sprintf(":%d", port))
 	return server, l, err
 }
 
 func launchTestQueries(t *testing.T, wg *sync.WaitGroup, c chan os.Signal) {
+	defer wg.Done()
 	var once sync.Once
-	for i := 0; i < 8; i++ {
+
+	for i := 0; i < concurrentRequestN; i++ {
+		wg.Add(1)
 		go runQuery(t, http.StatusOK, false, wg, &once)
 	}
 
@@ -101,121 +107,185 @@ func launchTestQueries(t *testing.T, wg *sync.WaitGroup, c chan os.Signal) {
 	c <- os.Interrupt
 	time.Sleep(waitTime)
 
-	for i := 0; i < 8; i++ {
+	for i := 0; i < concurrentRequestN; i++ {
+		wg.Add(1)
 		go runQuery(t, 0, true, wg, &once)
 	}
-
-	wg.Done()
 }
 
 func TestGracefulRun(t *testing.T) {
-	c := make(chan os.Signal, 1)
-
 	var wg sync.WaitGroup
-	wg.Add(1)
+	defer wg.Wait()
 
+	c := make(chan os.Signal, 1)
 	server, l, err := createListener(killTime / 2)
 	if err != nil {
 		t.Fatal(err)
 	}
 
+	wg.Add(1)
 	go func() {
+		defer wg.Done()
 		srv := &Server{Timeout: killTime, Server: server, interrupt: c}
 		srv.Serve(l)
-		wg.Done()
 	}()
 
 	wg.Add(1)
 	go launchTestQueries(t, &wg, c)
-	wg.Wait()
+}
+
+func TestGracefulRunLimitKeepAliveListener(t *testing.T) {
+	var wg sync.WaitGroup
+	defer wg.Wait()
+
+	c := make(chan os.Signal, 1)
+	server, l, err := createListener(killTime / 2)
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	wg.Add(1)
+	go func() {
+		defer wg.Done()
+		srv := &Server{
+			Timeout:      killTime,
+			ListenLimit:  concurrentRequestN,
+			TCPKeepAlive: 1 * time.Second,
+			Server:       server,
+			interrupt:    c,
+		}
+		srv.Serve(l)
+	}()
+
+	wg.Add(1)
+	go launchTestQueries(t, &wg, c)
 }
 
 func TestGracefulRunTimesOut(t *testing.T) {
-	c := make(chan os.Signal, 1)
-
 	var wg sync.WaitGroup
-	wg.Add(1)
+	defer wg.Wait()
 
+	c := make(chan os.Signal, 1)
 	server, l, err := createListener(killTime * 10)
 	if err != nil {
 		t.Fatal(err)
 	}
 
-	go func() {
-		srv := &Server{Timeout: killTime, Server: server, interrupt: c}
-		srv.Serve(l)
-		wg.Done()
-	}()
-
-	var once sync.Once
 	wg.Add(1)
 	go func() {
-		for i := 0; i < 8; i++ {
+		defer wg.Done()
+		srv := &Server{Timeout: killTime, Server: server, interrupt: c}
+		srv.Serve(l)
+	}()
+
+	wg.Add(1)
+	go func() {
+		defer wg.Done()
+		var once sync.Once
+
+		for i := 0; i < concurrentRequestN; i++ {
+			wg.Add(1)
 			go runQuery(t, 0, true, &wg, &once)
 		}
+
 		time.Sleep(waitTime)
 		c <- os.Interrupt
 		time.Sleep(waitTime)
-		for i := 0; i < 8; i++ {
+
+		for i := 0; i < concurrentRequestN; i++ {
+			wg.Add(1)
 			go runQuery(t, 0, true, &wg, &once)
 		}
-		wg.Done()
 	}()
-
-	wg.Wait()
-
 }
 
 func TestGracefulRunDoesntTimeOut(t *testing.T) {
-	c := make(chan os.Signal, 1)
-
 	var wg sync.WaitGroup
-	wg.Add(1)
+	defer wg.Wait()
 
+	c := make(chan os.Signal, 1)
 	server, l, err := createListener(killTime * 2)
 	if err != nil {
 		t.Fatal(err)
 	}
 
+	wg.Add(1)
 	go func() {
+		defer wg.Done()
 		srv := &Server{Timeout: 0, Server: server, interrupt: c}
 		srv.Serve(l)
-		wg.Done()
 	}()
 
 	wg.Add(1)
 	go launchTestQueries(t, &wg, c)
+}
+
+func TestGracefulRunDoesntTimeOutAfterConnectionCreated(t *testing.T) {
+	var wg sync.WaitGroup
+	defer wg.Wait()
+
+	c := make(chan os.Signal, 1)
+	server, l, err := createListener(killTime)
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	wg.Add(1)
+	go func() {
+		defer wg.Done()
+		srv := &Server{Timeout: 0, Server: server, interrupt: c}
+		srv.Serve(l)
+	}()
+	time.Sleep(waitTime)
+
+	// Make a sample first request. The connection will be left idle.
+	resp, err := http.Get(fmt.Sprintf("http://localhost:%d", port))
+	if err != nil {
+		panic(fmt.Sprintf("first request failed: %v", err))
+	}
+	resp.Body.Close()
+
+	wg.Add(1)
+	go func() {
+		defer wg.Done()
+
+		// With idle connections improperly handled, the server doesn't wait for this
+		// to complete and the request fails. It should be allowed to complete successfully.
+		_, err := http.Get(fmt.Sprintf("http://localhost:%d", port))
+		if err != nil {
+			t.Errorf("Get failed: %v", err)
+		}
+	}()
+
+	// Ensure the request goes out
+	time.Sleep(waitTime)
+	c <- os.Interrupt
 	wg.Wait()
 }
 
 func TestGracefulRunNoRequests(t *testing.T) {
-	c := make(chan os.Signal, 1)
-
 	var wg sync.WaitGroup
-	wg.Add(1)
+	defer wg.Wait()
 
+	c := make(chan os.Signal, 1)
 	server, l, err := createListener(killTime * 2)
 	if err != nil {
 		t.Fatal(err)
 	}
 
+	wg.Add(1)
 	go func() {
+		defer wg.Done()
 		srv := &Server{Timeout: 0, Server: server, interrupt: c}
 		srv.Serve(l)
-		wg.Done()
 	}()
 
 	c <- os.Interrupt
-
-	wg.Wait()
-
 }
 
 func TestGracefulForwardsConnState(t *testing.T) {
-	c := make(chan os.Signal, 1)
-	states := make(map[http.ConnState]int)
 	var stateLock sync.Mutex
-
+	states := make(map[http.ConnState]int)
 	connState := func(conn net.Conn, state http.ConnState) {
 		stateLock.Lock()
 		states[state]++
@@ -223,20 +293,23 @@ func TestGracefulForwardsConnState(t *testing.T) {
 	}
 
 	var wg sync.WaitGroup
-	wg.Add(1)
+	defer wg.Wait()
 
 	expected := map[http.ConnState]int{
-		http.StateNew:    8,
-		http.StateActive: 8,
-		http.StateClosed: 8,
+		http.StateNew:    concurrentRequestN,
+		http.StateActive: concurrentRequestN,
+		http.StateClosed: concurrentRequestN,
 	}
 
+	c := make(chan os.Signal, 1)
 	server, l, err := createListener(killTime / 2)
 	if err != nil {
 		t.Fatal(err)
 	}
 
+	wg.Add(1)
 	go func() {
+		defer wg.Done()
 		srv := &Server{
 			ConnState: connState,
 			Timeout:   killTime,
@@ -244,8 +317,6 @@ func TestGracefulForwardsConnState(t *testing.T) {
 			interrupt: c,
 		}
 		srv.Serve(l)
-
-		wg.Done()
 	}()
 
 	wg.Add(1)
@@ -305,7 +376,7 @@ func TestGracefulExplicitStopOverride(t *testing.T) {
 
 func TestBeforeShutdownAndShutdownInitiatedCallbacks(t *testing.T) {
 	var wg sync.WaitGroup
-	wg.Add(1)
+	defer wg.Wait()
 
 	server, l, err := createListener(1 * time.Millisecond)
 	if err != nil {
@@ -313,16 +384,18 @@ func TestBeforeShutdownAndShutdownInitiatedCallbacks(t *testing.T) {
 	}
 
 	beforeShutdownCalled := make(chan struct{})
-	cb1 := func() { close(beforeShutdownCalled) }
+	cb1 := func() bool { close(beforeShutdownCalled); return true }
 	shutdownInitiatedCalled := make(chan struct{})
 	cb2 := func() { close(shutdownInitiatedCalled) }
 
+	wg.Add(2)
 	srv := &Server{Server: server, BeforeShutdown: cb1, ShutdownInitiated: cb2}
 	go func() {
+		defer wg.Done()
 		srv.Serve(l)
-		wg.Done()
 	}()
 	go func() {
+		defer wg.Done()
 		time.Sleep(waitTime)
 		srv.Stop(killTime)
 	}()
@@ -348,6 +421,61 @@ func TestBeforeShutdownAndShutdownInitiatedCallbacks(t *testing.T) {
 	if !shutdownInitiated {
 		t.Fatal("shutdownInitiated should be true")
 	}
+}
+
+func TestBeforeShutdownCanceled(t *testing.T) {
+	var wg sync.WaitGroup
+	wg.Add(1)
+
+	server, l, err := createListener(1 * time.Millisecond)
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	beforeShutdownCalled := make(chan struct{})
+	cb1 := func() bool { close(beforeShutdownCalled); return false }
+	shutdownInitiatedCalled := make(chan struct{})
+	cb2 := func() { close(shutdownInitiatedCalled) }
+
+	srv := &Server{Server: server, BeforeShutdown: cb1, ShutdownInitiated: cb2}
+	go func() {
+		srv.Serve(l)
+		wg.Done()
+	}()
+	go func() {
+		time.Sleep(waitTime)
+		srv.Stop(killTime)
+	}()
+
+	beforeShutdown := false
+	shutdownInitiated := false
+	timeouted := false
+
+	for i := 0; i < 2; i++ {
+		select {
+		case <-beforeShutdownCalled:
+			beforeShutdownCalled = nil
+			beforeShutdown = true
+		case <-shutdownInitiatedCalled:
+			shutdownInitiatedCalled = nil
+			shutdownInitiated = true
+		case <-time.After(killTime):
+			timeouted = true
+		}
+	}
+
+	if !beforeShutdown {
+		t.Fatal("beforeShutdown should be true")
+	}
+	if !timeouted {
+		t.Fatal("timeouted should be true")
+	}
+	if shutdownInitiated {
+		t.Fatal("shutdownInitiated shouldn't be true")
+	}
+
+	srv.BeforeShutdown = func() bool { return true }
+	srv.Stop(killTime)
 
 	wg.Wait()
 }
@@ -367,17 +495,16 @@ func hijackingListener(srv *Server) (*http.Server, net.Listener, error) {
 		bufrw.Flush()
 	})
 
-	server := &http.Server{Addr: ":9654", Handler: mux}
-	l, err := net.Listen("tcp", ":9654")
+	server := &http.Server{Addr: fmt.Sprintf(":%d", port), Handler: mux}
+	l, err := net.Listen("tcp", fmt.Sprintf(":%d", port))
 	return server, l, err
 }
 
 func TestNotifyClosed(t *testing.T) {
-	c := make(chan os.Signal, 1)
-
 	var wg sync.WaitGroup
-	wg.Add(1)
+	defer wg.Wait()
 
+	c := make(chan os.Signal, 1)
 	srv := &Server{Timeout: killTime, interrupt: c}
 	server, l, err := hijackingListener(srv)
 	if err != nil {
@@ -386,13 +513,15 @@ func TestNotifyClosed(t *testing.T) {
 
 	srv.Server = server
 
+	wg.Add(1)
 	go func() {
+		defer wg.Done()
 		srv.Serve(l)
-		wg.Done()
 	}()
 
 	var once sync.Once
-	for i := 0; i < 8; i++ {
+	for i := 0; i < concurrentRequestN; i++ {
+		wg.Add(1)
 		runQuery(t, http.StatusOK, false, &wg, &once)
 	}
 
@@ -412,8 +541,10 @@ func TestNotifyClosed(t *testing.T) {
 }
 
 func TestStopDeadlock(t *testing.T) {
-	c := make(chan struct{})
+	var wg sync.WaitGroup
+	defer wg.Wait()
 
+	c := make(chan struct{})
 	server, l, err := createListener(1 * time.Millisecond)
 	if err != nil {
 		t.Fatal(err)
@@ -421,12 +552,14 @@ func TestStopDeadlock(t *testing.T) {
 
 	srv := &Server{Server: server, NoSignalHandling: true}
 
+	wg.Add(2)
 	go func() {
+		defer wg.Done()
 		time.Sleep(waitTime)
 		srv.Serve(l)
 	}()
-
 	go func() {
+		defer wg.Done()
 		srv.Stop(0)
 		close(c)
 	}()
@@ -525,6 +658,28 @@ func TestMultiInterrupts(t *testing.T) {
 	}
 }
 
+func TestLogFunc(t *testing.T) {
+	c := make(chan os.Signal, 1)
+
+	server, l, err := createListener(killTime * 10)
+	if err != nil {
+		t.Fatal(err)
+	}
+	var called bool
+	srv := &Server{Timeout: killTime, Server: server,
+		LogFunc: func(format string, args ...interface{}) {
+			called = true
+		}, interrupt: c}
+	stop := srv.StopChan()
+	go func() { srv.Serve(l) }()
+	c <- os.Interrupt
+	<-stop
+
+	if called != true {
+		t.Fatal("Expected LogFunc to be called.")
+	}
+}
+
 // SyncBuffer calls Done on the embedded wait group after each call to Write.
 type SyncBuffer struct {
 	*sync.WaitGroup
